# SUSE-specific operations
#
# Copyright 2008 Novell, Inc.
# Author: Wei Kong <wkong@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

from cStringIO import StringIO
import glob
import gzip
import os
import re
import shutil
import stat
import tempfile
import urlgrabber.grabber as grabber
import commands,sys,getopt,tempfile,shutil

from VMLinux import VMLinux
from exceptions import *
from log import log
import caps
import mount
import msg
import util
import vmdisks
import vminstall
import random

autoyast_vdev = vmdisks.get_possible_vdevs(False)[-1]

def tftp_success(statusoutput):
    errorre=re.compile("Error*")
    if errorre.match(statusoutput[1]):
        return False
    else:
        return True

def getServAddr():
    #Can't use dhclient, which will make network unstable
    #FIXME 1: is there any better method to get the dhcp server IP 
    #which doesn't effect network status
    #This method depend on udhcp-0.9.8-2usemac
    #udhcpc_command="/usr/bin/udhcpc"
    udhcpc_command="/sbin/udhcpc"
    udhcpc_script="/usr/share/udhcpc/outputpy.udhcp.sh"
    macraw = [ 0x00, 0x16, 0x3e, random.randint(0x00, 0x7f), 
            random.randint(0x00, 0xff), random.randint(0x00, 0xff) ]
    mac = ':'.join(map(lambda x: "%02x" % x, macraw))
    udhcp_result=commands.getstatusoutput(udhcpc_command+" -n -q -s "+
                                  udhcpc_script+" -M "+mac)
    if (udhcp_result[0] != 0):
        raise PXEError(PXEError.E_NO_SERVER)
    udhcplines=udhcp_result[1].split('\n')
    dhcpinfo={}
    for line in udhcplines:
        s = line.strip()
        if (s.find('=') != -1):
            f=s.split('=')
            dhcpinfo[f[0]]=f[1]

    sa=dhcpinfo['siaddr']
    log.debug('PXE Server address %s got' % sa)
    return sa

class VMPXE(VMLinux):
    global servaddr
    global kernel_list
    global initrd_list
    global pxe_label

    def __init__(self):
        self.__initialized = True

    pxe_label=''
    kernel_list={}
    initrd_list={}
	
    def getInstallImagesFromPath(self, root, progress=lambda f:None, supports_32nonpae=True):
        """
        ON PXE, use the getInstallImagesFromUrl()
        """
        return getInstallImagesFromUrl()

    def getPXEconfigFromPXEserver(self):
        try:
            servaddr=getServAddr()
        except Exception, e:
            raise e
        os.chdir(vminstall.paths.imagedir)
        commandstr="tftp "+servaddr+" -c get pxelinux.cfg/default"
        #commandstr="atftp -g -r pxelinux.cfg/default -l default 147.2.207.153" #for test
        cfgfilename="default"
        getpxeres=commands.getstatusoutput(commandstr)
        if not tftp_success(getpxeres) or not os.path.exists(cfgfilename):
    	    raise PXEError(PXEError.E_NO_CONFIG)
        cfgfile=open(cfgfilename)
        cfglines=cfgfile.readlines()
        # parse configure
        defaultrc=re.compile("default*")
        labelre=re.compile("label*")
        kernelre=re.compile("kernel*")
        appendre=re.compile("append*")
        initrdre=re.compile("initrd*")
        labelname='firstLabel'
        for line in cfglines:
            if (line[0]!='#'):
                line=line.strip()
                if (defaultrc.match(line)):
                    (k,v)=line.split()
                    pxe_label=v
                elif (labelre.match(line)):
                    (k,v)=line.split()
                    labelname=v
                elif (kernelre.match(line)):
                    (k,v)=line.split()
                    kernel_list[labelname]=v
                elif (appendre.match(line)):
                    for entry in line[6:].split():
                        if (entry.find('initrd=') != -1):
                            initrd_list[labelname]=entry[7:]
        cfgfile.close()
        return kernel_list

    def _getImageFromInstallArgs(self, options):
        #If not successful when get vmlinuz-xen initrd-xen from pxe-server:/tftpboot/, 
        #we could get vmlinuz-xen/initrd-xen from the INSTALL way 
        #which has been saved in getInstallArgs():/boot/{%arch}/vmlinuz-xen
        #FIXME 2:prase options._install_args
        return PXEError(PXEError.E_NO_KERNEL)

    def getInstallImagesFromUrl(self, options, url, progress=lambda f:None, supports_32nonpae=True):
        try:
            servaddr=getServAddr()
        except Exception, e:
            raise e
        os.chdir(vminstall.paths.imagedir)
        dlres={}
        log.debug('k/i: %s, %s'% (kernel_list[options.pxe_label], initrd_list[options.pxe_label]))
        for i in [kernel_list[options.pxe_label], initrd_list[options.pxe_label]]:
            cmd="tftp "+servaddr+" -c get "+ i
            #cmd="tftp 147.2.207.153 -c get "+ i #for test 
            dlres[i]=commands.getstatusoutput(cmd)
            if not tftp_success (dlres[i]):
               return _getImageFromInstallArgs(self, options)
        self.kfn = vminstall.paths.imagedir+kernel_list[options.pxe_label]
        self.ifn = vminstall.paths.imagedir+initrd_list[options.pxe_label]
        log.debug("Installation kernel/initrd: %s %s" % (self.kfn, self.ifn))
        return (self.kfn, self.ifn)

    def preprocessOptions(self, options):
        """Pulls settings out of extra_args."""
        if not options.extra_args:
            return
        args = options.extra_args.split()
        extra_args = ''
        for arg in args:
            if arg == 'TERM=xterm':
                # Discard.  Will be re-added later.
                pass
            elif arg.startswith('install='):
                if not options.source:
                    src = arg[8:]
                    if src.startswith('hd:/dev/'):
                        src = 'dev:/' + src[8:]
                    options.source = src
            elif arg.startswith('autoyast=') and not options.os_settings:
                options.os_settings = arg[9:]
            else:
                extra_args += arg + ' '
        options.extra_args = extra_args.strip()

    def setupOsSettings(self, guest, progress=lambda f:None):
        """Creates a temporary disk image, in which to pass the autoyast
        XML file."""
        ay = guest.options.os_settings
        if not ay or ay.startswith('ftp://') or ay.startswith('tftp://') or \
                     ay.startswith('http://') or ay.startswith('nfs://'):
            pass
        else:
            (fd, tmpdisk) = tempfile.mkstemp(prefix='autoyast.')
            disk = vmdisks.Disk(pdev=tmpdisk, vdev=autoyast_vdev, ro=True, blocks=2048,
                                full_virt=guest.options.full_virt, vmname=guest.options.name)
            try:
                disk.setup(progress=lambda f:progress(f*0.5))
                util.mke2fsImage(tmpdisk, progress=lambda f:progress(0.5+f*0.25))
                mounted_disk = mount.TemporaryLoopMount(tmpdisk)
                mountpoint = mounted_disk.mount(options='loop')
                progress(0.8)
                try:
                    mode = os.stat(ay).st_mode
                    if stat.S_ISREG(mode):
                        shutil.copyfile(ay, mountpoint + '/autoinst.xml')
                    elif stat.S_ISDIR(mode):
                        shutil.copyfile(ay + '/autoinst.xml', mountpoint + '/autoinst.xml')
                        shutil.copyfile(ay + '/info', mountpoint + '/info')
                finally:
                    progress(0.9)
                    mounted_disk.unmount()
            except:
                os.unlink(tmpdisk)
                raise
            guest.tmpdisks.append(disk)
        progress(1.0)

    def getKernelList(self):
        return kernel_list

    def setPXENumber(self, pNumber):
        pxe_label = pNumber[:]
    
    def getInstallArgs(self, options):
        pxeargs=""
        os.chdir(vminstall.paths.imagedir)
        cfgfilename="default"
        if not os.path.exists(cfgfilename):
            return None
        cfgfile=open(cfgfilename)
        labelre=re.compile('label '+options.pxe_label)
        startmatch = False
        log.debug('Label match to: label %s' % options.pxe_label)
        appendre=re.compile("append*")
        initrdre=re.compile("initrd*")
        cfglines=cfgfile.readlines()
        for line in cfglines:
            if (line[0]!='#'):
                line=line.strip()
                if (labelre.match(line)):
                    startmatch = True
                if startmatch and (appendre.match(line)):
                    for entry in line[6:].split():
                        if (entry.find('initrd=') != -1):
                            pass
                        else:
                            pxeargs += entry+' '
                    break
        cfgfile.close()
        log.debug('Return label:%s   args: %s' %(options.pxe_label, pxeargs))
        #save to options.extra_args also when PXE install
        options.extra_args=pxeargs[:]
        return pxeargs
